/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/ioqueue.h: Header for asynchronous I/O functions.
 */

/*
 * This header declares functionality for asynchronous I/O operations
 * (though no "output" functions are currently implemented due to lack of
 * need).  These are similar in concept to the POSIX "aio" library, but
 * also allow asynchronous open operations.
 *
 * To start an operation, call ioq_open() or ioq_read().  These functions
 * take the same parameters as the equivalent POSIX system calls, but
 * return an "I/O request ID" rather than a file descriptor or byte count.
 * This request ID can then be passed to ioq_poll() to check the status of
 * the request, or to ioq_wait() to wait for completion and obtain the
 * result.  Note that a pending request will continue to use system
 * resources even after the operation completes until its result has been
 * retrieved with ioq_wait().
 *
 * A pending request can be cancelled by calling ioq_cancel(), which will
 * stop the operation (if possible) and set the request to an error state.
 * the request must still be waited for with ioq_wait().
 *
 * File descriptors opened with ioq_open() can be used with regular read(),
 * and likewise for open() and ioq_read(); however, behavior is undefined
 * if read() is called while an asynchronous read is pending on the same
 * file descriptor.  There is no asynchronous close operation, so file
 * descriptors from ioq_open() should be closed with regular close().
 *
 * ioq_*() functions set errno on error, like regular library functions.
 * However, the ESRCH error code is used to indicate that an invalid
 * request ID was passed to a function (like EBADFD for file descriptors).
 *
 * All ioq_*() functions except ioq_reset() are thread-safe.
 *
 * Note that ioq_open() is not currently used by posix/files.c.
 */

#ifndef SIL_SRC_SYSDEP_POSIX_IOQUEUE_H
#define SIL_SRC_SYSDEP_POSIX_IOQUEUE_H

#include <sys/types.h>  // For off_t definition.

/*************************************************************************/
/*************************************************************************/

/**
 * ioq_open:  Start an asynchronous open operation.
 *
 * [Parameters]
 *     path: Path of file to open.
 *     flags: Open flags (must not include O_WRONLY or O_RDWR).
 * [Return value]
 *     I/O request ID (nonzero), or zero on error.
 * [Operation result]
 *     New file descriptor, or -1 on error.
 */
extern int ioq_open(const char *path, int flags);

/**
 * ioq_read:  Start an asynchronous read operation.
 *
 * [Parameters]
 *     fd: File descriptor to read from.
 *     buf: Buffer into which to read.
 *     count: Number of bytes to read.
 *     pos: File position for read.
 * [Return value]
 *     I/O request ID (nonzero), or zero on error.
 * [Operation result]
 *     Number of bytes read, or -1 on error.  End-of-file is not considered
 *     an error.
 */
extern int ioq_read(int fd, void *buf, ssize_t count, off_t pos);

/**
 * ioq_poll:  Return the completion status of an asynchronous operation.
 *
 * [Parameters]
 *     request: I/O request ID.
 * [Return value]
 *     True if the request has completed or the request ID is invalid;
 *     false if the request is still in progress.
 */
extern int ioq_poll(int request);

/**
 * ioq_wait:  Wait for an asynchronous operation to complete, and return
 * its result.  After calling this function, the request ID is no longer
 * valid.
 *
 * [Parameters]
 *     request: I/O request ID.
 *     error_ret: Pointer to variable to receive the error code for a
 *         failed operation, 0 for a successful operation.  May be NULL if
 *         the error code is not needed.
 * [Return value]
 *     Operation result, or -1 if the request ID is invalid.
 */
extern ssize_t ioq_wait(int request, int *error_ret);

/**
 * ioq_cancel:  Cancel an asynchronous operation.  The request must still
 * be waited for with ioq_wait(), which will return failure with error
 * ECANCELED.
 *
 * Note that while this function always succeeds for a valid request ID,
 * it will not necessarily abort the I/O itself, depending on the state of
 * the operation.  However, aborting an open operation will free the file
 * descriptor if the open succeeded.
 *
 * [Parameters]
 *     request: I/O request ID.
 */
extern void ioq_cancel(int request);

/**
 * ioq_reset:  Cancel and invalidate all asynchronous operations which have
 * not yet been waited for, as though ioq_cancel() and ioq_wait() had been
 * called on each one; then reset all internal state.
 *
 * This function is provided primarily for testing purposes.  Use with
 * caution -- if any references to pending operations are held, they may
 * interfere with future operations due to request ID collisions.
 *
 * [Parameters]
 *     preallocate: True to preallocate MAX_ASYNC_READS I/O slots and start the
 *         I/O thread; false to free all slots and leave the thread stopped.
 */
extern void ioq_reset(int preallocate);

/*************************************************************************/
/************************ Test control interface *************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

/**
 * TEST_posix_ioqueue_move_on_realloc:  Enable or disable moving the I/O
 * request array on expansion.  When enabled, the I/O request array will
 * always be moved to a different address when it is expanded, to expose
 * errors caused by dangling pointers.
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_move_on_realloc(int enable);

/**
 * TEST_posix_ioqueue_block_io_thread:  Enable or disable blocking of the
 * I/O thread.  While enabled, the I/O thread will stop executing; this can
 * be used to intentionally queue up several requests before allowing any
 * to be processed.
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_block_io_thread(int enable);

/**
 * TEST_posix_ioqueue_unblock_on_wait:  Enable or disable unblocking of the
 * I/O thread on calls to ioq_wait().  Has no effect if the I/O thread is
 * not blocked via TEST_posix_ioqueue_block_io_thread().
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_unblock_on_wait(int enable);

/**
 * TEST_posix_ioqueue_permfail_next_read:  Enable or disable force-failing
 * of the next call to ioq_read() with ENOMEM.  If enabled, this only takes
 * effect for one call and is subsequently disabled.
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_permfail_next_read(int enable);

/**
 * TEST_posix_ioqueue_tempfail_next_read:  Enable or disable force-failing
 * of the next call to ioq_read() with EAGAIN.  If enabled, this only takes
 * effect for one call and is subsequently disabled.
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_tempfail_next_read(int enable);

/**
 * TEST_posix_ioqueue_iofail_next_read:  Enable or disable force-failing of
 * the next read operation enqueued by ioq_read() with EIO.  If enabled,
 * this only takes effect for one operation and is subsequently disabled.
 *
 * [Parameters]
 *     enable: True to enable, false to disable.
 */
extern void TEST_posix_ioqueue_iofail_next_read(int enable);

#endif  // INCLUDE_TESTS

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_POSIX_IOQUEUE_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
